# Copyright 2019 Novartis Institutes for BioMedical Research Inc. Licensed
# under the Apache License, Version 2.0 (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy
# of the License at http://www.apache.org/licenses/LICENSE-2.0. Unless
# required by applicable law or agreed to in writing, software distributed
# under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES
# OR CONDITIONS OF ANY KIND, either express or implied. See the License for
# the specific language governing permissions and limitations under the License.

import os

from flask_api import status

from cellxgene_gateway import env
from cellxgene_gateway.cellxgene_exception import CellxgeneException
from cellxgene_gateway.dir_util import make_h5ad
from cellxgene_gateway.cache_key import CacheKey

def get_key(path):
    if path == "/" or path == "":
        raise CellxgeneException(
            "No matching dataset found.", status.HTTP_404_NOT_FOUND
        )

    trimmed = path[:-1] if path[-1] == "/" else path
    try:
        # valid paths come in three forms:
        if trimmed.endswith('.h5ad') and data_file_exists(trimmed):
            # 1) somedir/dataset.h5ad: a dataset
            return CacheKey(trimmed, trimmed, None)
        elif trimmed.endswith('.csv'):

            # 2) somedir/dataset_annotations/saldaal1-T5HMVBNV.csv : an actual annotations file.
            annotations_dir = os.path.split(trimmed)[0]
            dataset = make_h5ad(annotations_dir)
            if data_file_exists(dataset):
                data_dir_ensure(annotations_dir)
                return CacheKey(trimmed, dataset, trimmed)
        elif trimmed.endswith('_annotations') and data_dir_exists(trimmed):
            # 3) somedir/dataset_annotations: an annotation directory. The corresponding h5ad must exist, but the directory may not.
            dataset = make_h5ad(trimmed)
            if data_file_exists(dataset):
                return CacheKey(trimmed, dataset, '')
    except CellxgeneException:
        pass
    split = os.path.split(trimmed)
    return get_key(split[0])

def validate_exists(file_path):
    if not os.path.exists(file_path):
        raise CellxgeneException(
            "File does not exist: " + file_path, status.HTTP_400_BAD_REQUEST
        )

def validate_is_file(file_path):
    validate_exists(file_path)
    if not os.path.isfile(file_path):
        raise CellxgeneException(
            "Path is not file: " + file_path, status.HTTP_400_BAD_REQUEST
        )
    return
def validate_is_dir(file_path):
    validate_exists(file_path)
    if not os.path.isdir(file_path):
        raise CellxgeneException(
            "Path is not dir: " + file_path, status.HTTP_400_BAD_REQUEST
        )
    return

def data_file_exists(dataset):
    file_path = os.path.join(env.cellxgene_data, dataset)
    validate_is_file(file_path)
    return True
def data_dir_exists(dataset):
    file_path = os.path.join(env.cellxgene_data, dataset)
    validate_is_dir(file_path)
    return True
def data_dir_ensure(dataset):
    file_path = os.path.join(env.cellxgene_data, dataset)
    if not os.path.exists(file_path):
        os.makedirs(file_path)

def get_file_path(key):
    dataset = key.dataset
    file_path = os.path.join(env.cellxgene_data, dataset)
    validate_is_file(file_path)
    return file_path

def get_annotation_file_path(key):
    if key.annotation_file is None:
        return None
    if key.annotation_file == '':
        return ''
    file_path = os.path.join(env.cellxgene_data, key.annotation_file)
    return file_path



# Copyright 2019 Novartis Institutes for BioMedical Research Inc. Licensed
# under the Apache License, Version 2.0 (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy
# of the License at http://www.apache.org/licenses/LICENSE-2.0. Unless
# required by applicable law or agreed to in writing, software distributed
# under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES
# OR CONDITIONS OF ANY KIND, either express or implied. See the License for
# the specific language governing permissions and limitations under the License.

import os

from flask_api import status

from cellxgene_gateway import env
from cellxgene_gateway.cellxgene_exception import CellxgeneException

# There are three kinds of CacheKey:
# 1) somedir/dataset.h5ad: a dataset
#    in this case, pathpart == dataset == 'somedir/dataset.h5ad'
# 2) somedir/dataset_annotations/saldaal1-T5HMVBNV.csv : an actual annotaitons file.
#    in this case, pathpart == 'dataset_annotations/saldaal1-T5HMVBNV.csv', dataset == 'somedir/dataset.h5ad'
# 3) somedir/dataset_annotations: an annotation directory. The corresponding h5ad must exist, but the directory may not.
#    in this case, pathpart == 'dataset_annotations', dataset == 'somedir/dataset.h5ad'

class CacheKey:
    def __init__(self, pathpart, dataset, annotation_file):
        self.pathpart = pathpart
        self.dataset = dataset
        self.annotation_file = annotation_file



# Copyright 2019 Novartis Institutes for BioMedical Research Inc. Licensed
# under the Apache License, Version 2.0 (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy
# of the License at http://www.apache.org/licenses/LICENSE-2.0. Unless
# required by applicable law or agreed to in writing, software distributed
# under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES
# OR CONDITIONS OF ANY KIND, either express or implied. See the License for
# the specific language governing permissions and limitations under the License.

from cellxgene_gateway import env
from json import loads


def get_extra_scripts():
    # can be array of script tags to inject on every page, e.g. for google analytics could be
    # ['https://www.googletagmanager.com/gtag/js?id=UA-123456-2',
    #  f"{env.external_protocol}://{env.external_host}/static/js/google_ua.js"]
    # where google_ua.js is a script you add to the static/js folder prior to deployment.
    return [] if env.extra_scripts is None else loads(env.extra_scripts)



# Copyright 2019 Novartis Institutes for BioMedical Research Inc. Licensed
# under the Apache License, Version 2.0 (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy
# of the License at http://www.apache.org/licenses/LICENSE-2.0. Unless
# required by applicable law or agreed to in writing, software distributed
# under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES
# OR CONDITIONS OF ANY KIND, either express or implied. See the License for
# the specific language governing permissions and limitations under the License.
import psutil
import logging
import datetime

from flask import make_response, request, render_template
from requests import get, post, put

from cellxgene_gateway import env
from cellxgene_gateway.cellxgene_exception import CellxgeneException
from cellxgene_gateway.util import current_time_stamp
from cellxgene_gateway.flask_util import querystring

class CacheEntry:
    def __init__(
        self,
        pid,
        key,
        port,
        launchtime,
        timestamp,
        status,
        message,
        all_output,
        stderr,
        http_status,
    ):
        self.pid = pid
        self.key = key
        self.port = port
        self.launchtime = launchtime
        self.timestamp = timestamp
        self.status = status
        self.message = message
        self.all_output = all_output
        self.stderr = stderr
        self.http_status = http_status

    @classmethod
    def for_key(cls, key, port):

        return cls(
            None,
            key,
            port,
            current_time_stamp(),
            current_time_stamp(),
            "loading",
            None,
            None,
            None,
            None,
        )

    def set_loaded(self, pid):
        self.pid = pid
        self.status = "loaded"

    def set_error(self, message, stderr, http_status):
        self.message = message
        self.stderr = stderr
        self.http_status = http_status
        self.status = "error"

    def append_output(self, output):
        if self.all_output == None:
            self.all_output = output
        else:
            self.all_output += output

    def terminate(self):
        pid = self.pid
        if pid != None and self.status != "terminated":
            terminated = []
            def on_terminate(p):
                terminated.append(p.pid)
            p = psutil.Process(pid)
            children = p.children()
            for child in children:
                child.terminate()
            psutil.wait_procs(children, callback=on_terminate)
            terminated.append(p.pid)
            p.terminate()
            psutil.wait_procs([p], callback=on_terminate)
            logging.getLogger("cellxgene_gateway").info(f"terminated {terminated}")
        self.status = "terminated"

    def serve_content(self, path):
        gateway_basepath = (
            f"{env.external_protocol}://{env.external_host}/view/{self.key.pathpart}/"
        )
        subpath = path[len(self.key.pathpart) :]  # noqa: E203
            
        if len(subpath) == 0:
            r = make_response(f"Redirect to {gateway_basepath}\n", 301)
            r.headers["location"] = gateway_basepath+querystring()
            return r
        elif self.status == "loading":
            launch_time = datetime.datetime.fromtimestamp(self.launchtime)
            return render_template(
                "loading.html", launchtime=launch_time, all_output=self.all_output
            )

        port = self.port
        cellxgene_basepath = f"http://127.0.0.1:{port}"
        headers = {}
        copy_headers = [
            'accept',
            'accept-encoding',
            'accept-language',
            'cache-control',
            'connection',
            'content-length',
            'content-type',
            'cookie',
            'host',
            'origin',
            'pragma',
            'referer',
            'sec-fetch-mode',
            'sec-fetch-site',
            'user-agent'
        ]
        for h in copy_headers:
            if h in request.headers:
                headers[h] = request.headers[h]

        full_path = cellxgene_basepath + subpath + querystring()

        if request.method in ["GET", "HEAD", "OPTIONS"]:
            cellxgene_response = get(
                full_path, headers=headers
            )
        elif request.method == "PUT":
            cellxgene_response = put(
                full_path,
                headers=headers,
                data=request.data,
            )
        elif request.method == "POST":
            cellxgene_response = post(
                full_path,
                headers=headers,
                data=request.data,
            )
        else:
            raise CellxgeneException(
                f"Unexpected method {request.method}", 400
            )
        content_type = cellxgene_response.headers["content-type"]
        if "text" in content_type:
            cellxgene_content = cellxgene_response.content.decode()
            gateway_content = cellxgene_content.replace(
                "http://fonts.gstatic.com", "https://fonts.gstatic.com"
            ).replace(cellxgene_basepath, gateway_basepath)
        else:
            gateway_content = cellxgene_response.content

        resp_headers = {}
        for h in copy_headers:
            if h in cellxgene_response.headers:
                resp_headers[h] = cellxgene_response.headers[h]

        gateway_response = make_response(
            gateway_content,
            cellxgene_response.status_code,
            resp_headers,
        )

        return gateway_response



# Copyright 2019 Novartis Institutes for BioMedical Research Inc. Licensed
# under the Apache License, Version 2.0 (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy
# of the License at http://www.apache.org/licenses/LICENSE-2.0. Unless
# required by applicable law or agreed to in writing, software distributed
# under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES
# OR CONDITIONS OF ANY KIND, either express or implied. See the License for
# the specific language governing permissions and limitations under the License.

# import BaseHTTPServer
import os
import logging
from threading import Thread, Lock
import json

from flask import (
    Flask,
    redirect,
    make_response,
    render_template,
    request,
    send_from_directory,
    url_for,
)
from flask_api import status
from werkzeug.utils import secure_filename

from cellxgene_gateway import env
from cellxgene_gateway.backend_cache import BackendCache
from cellxgene_gateway.cellxgene_exception import CellxgeneException
from cellxgene_gateway.dir_util import create_dir, is_subdir
from cellxgene_gateway.filecrawl import recurse_dir, render_entries
from cellxgene_gateway.extra_scripts import get_extra_scripts
from cellxgene_gateway.process_exception import ProcessException
from cellxgene_gateway.prune_process_cache import PruneProcessCache
from cellxgene_gateway.util import current_time_stamp
from cellxgene_gateway.path_util import get_key

app = Flask(__name__)

def _force_https(app):
    def wrapper(environ, start_response):
        environ['wsgi.url_scheme'] = env.external_protocol
        return app(environ, start_response)
    return wrapper
app.wsgi_app = _force_https(app.wsgi_app)

cache = BackendCache()
location = f"{env.external_protocol}://{env.external_host}"


@app.errorhandler(CellxgeneException)
def handle_invalid_usage(error):

    message = f"{error.http_status} Error : {error.message}"

    return (
        render_template(
            "cellxgene_error.html",
            extra_scripts=get_extra_scripts(),
            message=message,
        ),
        error.http_status,
    )


@app.errorhandler(ProcessException)
def handle_invalid_process(error):

    message = []

    message.append(error.message)
    message.append(f"{error.http_status} Error.")
    message.append(f"Stdout: {error.stdout}")
    message.append(f"Stderr: {error.stderr}")

    return (
        render_template(
            "process_error.html",
            extra_scripts=get_extra_scripts(),
            message=error.message,
            http_status=error.http_status,
            stdout=error.stdout,
            stderr=error.stderr,
            dataset=error.key.dataset,
            annotation_file=error.key.annotation_file,
        ),
        error.http_status,
    )


@app.route("/favicon.ico")
def favicon():
    return send_from_directory(
        os.path.join(app.root_path, "static"),
        "nibr.ico",
        mimetype="image/vnd.microsof.icon",
    )


@app.route("/")
def index():
    users = [
        name
        for name in os.listdir(env.cellxgene_data)
        if os.path.isdir(os.path.join(env.cellxgene_data, name))
    ]
    return render_template(
        "index.html",
        ip=env.ip,
        cellxgene_data=env.cellxgene_data,
        extra_scripts=get_extra_scripts(),
        users=users,
        enable_upload=env.enable_upload,
    )

def make_user():
    dir_name = request.form["directory"]

    create_dir(env.cellxgene_data, dir_name)

    return redirect(location, code=302)


def make_subdir():
    parent_path = os.path.join(env.cellxgene_data, request.form["usernames"])
    dir_name = request.form["directory"]

    create_dir(parent_path, dir_name)

    return redirect(location, code=302)


def upload_file():
    upload_dir = request.form["path"]

    full_upload_path = os.path.join(env.cellxgene_data, upload_dir)
    if is_subdir(full_upload_path, env.cellxgene_data) and os.path.isdir(full_upload_path):
        if request.method == "POST":
            if "file" in request.files:
                f = request.files["file"]
                if f and f.filename.endswith(".h5ad"):
                    f.save(
                        os.path.join(full_upload_path, secure_filename(f.filename))
                    )
                    return redirect("/filecrawl.html", code=302)
                else:
                    raise CellxgeneException(
                        "Uploaded file must be in anndata (.h5ad) format.",
                        status.HTTP_400_BAD_REQUEST,
                    )
            else:
                raise CellxgeneException(
                    "A file must be chosen to upload.",
                    status.HTTP_400_BAD_REQUEST,
                )
    else:
        raise CellxgeneException(
            "Invalid directory.", status.HTTP_400_BAD_REQUEST
        )

    return redirect(location, code=302)


if env.enable_upload:
    app.add_url_rule('/make_user', 'make_user', make_user, methods=["POST"])
    app.add_url_rule('/make_subdir', 'make_subdir', make_subdir, methods=["POST"])
    app.add_url_rule('/upload_file', 'upload_file', upload_file, methods=["POST"])

@app.route("/filecrawl.html")
def filecrawl():
    entries = recurse_dir(env.cellxgene_data)
    rendered_html = render_entries(entries)
    resp = make_response(render_template(
        "filecrawl.html",
        extra_scripts=get_extra_scripts(),
        rendered_html=rendered_html,
    ))
    resp.headers["Cache-Control"] = "no-cache, no-store, must-revalidate"
    resp.headers["Pragma"] = "no-cache"
    resp.headers["Expires"] = "0"
    resp.headers['Cache-Control'] = 'public, max-age=0'
    return resp

@app.route("/filecrawl/<path:path>")
def do_filecrawl(path):
    filecrawl_path = os.path.join(env.cellxgene_data, path)
    if not os.path.isdir(filecrawl_path):
        raise CellxgeneException(
            "Path is not directory: " + filecrawl_path, status.HTTP_400_BAD_REQUEST
        )
    entries = recurse_dir(filecrawl_path)
    rendered_html = render_entries(entries)
    return render_template(
        "filecrawl.html",
        extra_scripts=get_extra_scripts(),
        rendered_html=rendered_html,
        path=path,
    )

entry_lock = Lock()
@app.route("/view/<path:path>", methods=["GET", "PUT", "POST"])
def do_view(path):
    key = get_key(path)
    print(f"view path={path}, dataset={key.dataset}, annotation_file= {key.annotation_file}, key={key.pathpart}")
    with entry_lock:
        match = cache.check_entry(key)
        if match is None:
            uascripts = get_extra_scripts()
            match = cache.create_entry(key, uascripts)

    match.timestamp = current_time_stamp()

    if match.status == "loaded" or match.status == "loading":
        return match.serve_content(path)
    elif match.status == "error":
        raise ProcessException.from_cache_entry(match)


@app.route("/cache_status", methods=["GET"])
def do_GET_status():
    return render_template("cache_status.html", entry_list=cache.entry_list)

@app.route("/cache_status.json", methods=["GET"])
def do_GET_status_json():
    return json.dumps({'launchtime':app.launchtime,
    'entry_list':[{
        'dataset': entry.key.dataset,
        'annotation_file': entry.key.annotation_file,
        'launchtime': entry.launchtime,
        'last_access': entry.timestamp,
        'status': entry.status
    } for entry in cache.entry_list]})

@app.route("/relaunch/<path:path>", methods=["GET"])
def do_relaunch(path):
    key = get_key(path)
    match = cache.check_entry(key)
    if not match is None:
        match.terminate()
    qs = request.query_string.decode()
    return redirect(url_for("do_view", path=path) + (f'?{qs}' if len(qs) > 0 else ''), code=302)

@app.route("/terminate/<path:path>", methods=["GET"])
def do_terminate(path):
    key = get_key(path)
    match = cache.check_entry(key)
    if not match is None:
        match.terminate()
    return redirect(url_for("do_GET_status"), code=302)


def main():
    logging.basicConfig(level=logging.INFO, format='%(asctime)s:%(name)s:%(levelname)s:%(message)s')
    env.validate()
    pruner = PruneProcessCache(cache)

    background_thread = Thread(target=pruner)
    background_thread.start()

    app.launchtime = current_time_stamp()
    app.run(host="0.0.0.0", port=env.gateway_port, debug=False)


if __name__ == "__main__":
    main()



# Copyright 2019 Novartis Institutes for BioMedical Research Inc. Licensed
# under the Apache License, Version 2.0 (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy
# of the License at http://www.apache.org/licenses/LICENSE-2.0. Unless
# required by applicable law or agreed to in writing, software distributed
# under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES
# OR CONDITIONS OF ANY KIND, either express or implied. See the License for
# the specific language governing permissions and limitations under the License.

import os
from cellxgene_gateway import env
from cellxgene_gateway.dir_util import make_h5ad, make_annotations, annotations_suffix

def recurse_dir(path):
    if not os.path.exists(path):
        raise CellxgeneException(
            "The given path does not exist.", status.HTTP_400_BAD_REQUEST
        )

    all_entries = sorted(os.listdir(path))
    def is_h5ad(el):
        return el.endswith('.h5ad') and os.path.isfile(os.path.join(path, el))
    h5ad_entries = [x for x in all_entries if is_h5ad(x)]
    annotation_dir_entries = [x for x in all_entries if x.endswith(annotations_suffix) and make_h5ad(x) in h5ad_entries]
    def list_annotations(el):
        full_path = os.path.join(path, el)
        if not os.path.isdir(full_path):
            entries = []
        else:
            entries = [{
                "name": x[:-13] if (len(x) > 13 and x[-13] in ['-','_']) else (
                    x[:-4] if x.endswith('.csv') else x),
                "path": os.path.join(full_path, x).replace(env.cellxgene_data, ""),
            } for x in sorted(os.listdir(full_path)) if x.endswith('.csv') and os.path.isfile(os.path.join(full_path, x))]
        return [{"name":'new', "class":'new', "path":full_path.replace(env.cellxgene_data, "")}] + entries

    def make_entry(el):
        full_path = os.path.join(path, el)
        if el in h5ad_entries:
            return {
                "path": full_path.replace(env.cellxgene_data, ""),
                "name": el,
                "type": "file",
                "annotations": list_annotations(make_annotations(el)),
            }
        elif os.path.isdir(full_path) and el not in annotation_dir_entries:
            return {
                "path": full_path.replace(env.cellxgene_data, ""),
                "name": el,
                "type": "directory",
                "children": recurse_dir(full_path),
            }
        else:
            return {
                "path": full_path,
                "name": el,
                "type": "neither",
            }

    return [make_entry(x) for x in all_entries]


def render_entries(entries):
    return "<ul>" + "\n".join([render_entry(e) for e in entries]) + "</ul>"

def get_url(entry):
    return f"/view/{ entry['path'].lstrip('/') }"
def get_class(entry):
    return f" class='{entry['class']}'" if 'class' in entry else ''

def render_annotations(entry):
    if len(entry['annotations']) > 0:
        return ' | annotations: ' + ", ".join([f"<a href='{get_url(a)}'{get_class(a)}>{a['name']}</a>" for a in entry['annotations']])
    else:
        return ''

def render_entry(entry):
    if entry["type"] == "file":
        return f"<li> <a href='{ get_url(entry) }'>{entry['name']}</a> {render_annotations(entry)}</li>"
    elif entry["type"] == "directory":
        url = f"/filecrawl/{entry['path'].lstrip('/')}"
        return f"<li><a href='{url}'>{entry['name']}</a>{render_entries(entry['children'])}</li>"
    else:
        return ""



# Copyright 2019 Novartis Institutes for BioMedical Research Inc. Licensed
# under the Apache License, Version 2.0 (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy
# of the License at http://www.apache.org/licenses/LICENSE-2.0. Unless
# required by applicable law or agreed to in writing, software distributed
# under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES
# OR CONDITIONS OF ANY KIND, either express or implied. See the License for
# the specific language governing permissions and limitations under the License.

import logging
import subprocess

from flask_api import status
from cellxgene_gateway.env import enable_annotations, enable_backed_mode
from cellxgene_gateway.process_exception import ProcessException
from cellxgene_gateway.dir_util import make_annotations
from cellxgene_gateway.path_util import get_file_path, get_annotation_file_path


class SubprocessBackend:
    def __init__(self):
        pass

    def create_cmd(self, cellxgene_loc, file_path, port, scripts, annotation_file_path):
        if enable_annotations and not annotation_file_path is None:
            if annotation_file_path == "":
                extra_args = f" --annotations-dir {make_annotations(file_path)}"
            else:
                extra_args = f" --annotations-file {annotation_file_path}"
        else:
            extra_args = " --disable-annotations"
        if enable_backed_mode:
            extra_args += " --backed"
        cmd = (
            f"yes | {cellxgene_loc} launch {file_path}"
            + " --port "
            + str(port)
            + " --host 127.0.0.1"
            + extra_args
        )

        for s in scripts:
            cmd += f" --scripts {s}"

        return cmd

    def launch(self, cellxgene_loc, scripts, cache_entry):

        cmd = self.create_cmd(
            cellxgene_loc, get_file_path(cache_entry.key), cache_entry.port, scripts, get_annotation_file_path(cache_entry.key)
        )
        logging.getLogger("cellxgene_gateway").info(f"launching {cmd}")
        process = subprocess.Popen(
            [cmd], stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True
        )

        while True:
            output = process.stdout.readline().decode()
            if output == "[cellxgene] Type CTRL-C at any time to exit.\n":
                break
            elif output == "":
                stderr = process.stderr.read().decode()
                if (
                    "Error while loading file" in stderr
                    or "Could not open file" in stderr
                ):
                    message = "File was invalid."
                    http_status = status.HTTP_400_BAD_REQUEST
                else:
                    message = "Cellxgene failed to launch dataset."
                    http_status = status.HTTP_500_INTERNAL_SERVER_ERROR

                cache_entry.status = "error"
                cache_entry.set_error(message, stderr, http_status)

                raise ProcessException.from_cache_entry(cache_entry)
            else:
                cache_entry.append_output(output)

        cache_entry.set_loaded(process.pid)

        return



# Copyright 2019 Novartis Institutes for BioMedical Research Inc. Licensed
# under the Apache License, Version 2.0 (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy
# of the License at http://www.apache.org/licenses/LICENSE-2.0. Unless
# required by applicable law or agreed to in writing, software distributed
# under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES
# OR CONDITIONS OF ANY KIND, either express or implied. See the License for
# the specific language governing permissions and limitations under the License.

import os

from flask_api import status

from cellxgene_gateway import env
from cellxgene_gateway.cellxgene_exception import CellxgeneException


def is_subdir(full_path, parent_path):
    subdir = os.path.realpath(full_path)
    parent = os.path.realpath(parent_path)
    return subdir.startswith(parent)


def create_dir(parent_path, dir_name):
    full_path = os.path.join(parent_path, dir_name)

    if "/" in dir_name:
        raise CellxgeneException(
            "Please have no slashes in the intended directory.",
            status.HTTP_400_BAD_REQUEST,
        )
    elif not os.path.exists(parent_path):
        raise CellxgeneException(
            "The selected User directory does not exist.",
            status.HTTP_400_BAD_REQUEST,
        )
    elif os.path.exists(full_path):
        raise CellxgeneException(
            "The provided subdirectory already exists within Directory.",
            status.HTTP_400_BAD_REQUEST,
        )
    elif not is_subdir(full_path, parent_path):
        raise CellxgeneException(
            "The directory must be a subdirectory of the parent path.",
            status.HTTP_400_BAD_REQUEST,
        )
    elif not os.path.isdir(parent_path):
        raise CellxgeneException(
            "The parent is not a directory.", status.HTTP_400_BAD_REQUEST
        )
    else:
        os.mkdir(full_path)

annotations_suffix = '_annotations'
def make_h5ad(el):
    return el[:-len(annotations_suffix)]+'.h5ad'
def make_annotations(el):
    return el[:-5]+annotations_suffix



# Copyright 2019 Novartis Institutes for BioMedical Research Inc. Licensed
# under the Apache License, Version 2.0 (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy
# of the License at http://www.apache.org/licenses/LICENSE-2.0. Unless
# required by applicable law or agreed to in writing, software distributed
# under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES
# OR CONDITIONS OF ANY KIND, either express or implied. See the License for
# the specific language governing permissions and limitations under the License.

import time
from threading import Thread

from flask_api import status

from cellxgene_gateway import env
from cellxgene_gateway.cache_entry import CacheEntry
from cellxgene_gateway.cellxgene_exception import CellxgeneException
from cellxgene_gateway.subprocess_backend import SubprocessBackend

process_backend = SubprocessBackend()


def is_port_in_use(port):
    import socket
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        return s.connect_ex(('localhost', port)) == 0

class BackendCache:
    def __init__(self):
        self.entry_list = []

    def get_ports(self):
        contents = self.entry_list
        return [c.port for c in contents]

    def check_entry(self, key):
        contents = self.entry_list
        matches = [
            c
            for c in contents
            if c.key.dataset == key.dataset and c.key.annotation_file == key.annotation_file and c.status != "terminated"
        ]

        if len(matches) == 0:
            return None
        elif len(matches) == 1:
            return matches[0]
        else:
            raise CellxgeneException(
                status.HTTP_500_INTERNAL_SERVER_ERROR,
                "Found " + str(len(matches)) + " for " + dataset,
            )

    def create_entry(self, key, scripts):
        port = 8000
        existing_ports = self.get_ports()

        while (port in existing_ports) or is_port_in_use(port):
            port += 1

        entry = CacheEntry.for_key(key, port)

        background_thread = Thread(
            target=process_backend.launch,
            args=(env.cellxgene_location, scripts, entry),
        )
        background_thread.start()

        self.entry_list.append(entry)

        time.sleep(1)  # Automatic refresh is too fast, needs a second to pause

        return entry

    def prune(self, process):
        self.entry_list.remove(process)
        process.terminate()



# Copyright 2019 Novartis Institutes for BioMedical Research Inc. Licensed
# under the Apache License, Version 2.0 (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy
# of the License at http://www.apache.org/licenses/LICENSE-2.0. Unless
# required by applicable law or agreed to in writing, software distributed
# under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES
# OR CONDITIONS OF ANY KIND, either express or implied. See the License for
# the specific language governing permissions and limitations under the License.

import os
import logging
import socket

cellxgene_location = os.environ.get("CELLXGENE_LOCATION")
cellxgene_data = os.environ.get("CELLXGENE_DATA")
gateway_port = int(os.environ.get("GATEWAY_PORT", "5005"))
external_host = os.environ.get("EXTERNAL_HOST", os.environ.get("GATEWAY_HOST", f"localhost:{gateway_port}"))
external_protocol = os.environ.get("EXTERNAL_PROTOCOL", os.environ.get("GATEWAY_PROTOCOL", "http"))
ip = os.environ.get("GATEWAY_IP")
extra_scripts = os.environ.get("GATEWAY_EXTRA_SCRIPTS")
ttl = os.environ.get("GATEWAY_TTL")
enable_upload = os.environ.get("GATEWAY_ENABLE_UPLOAD", "").lower() in ['true', '1']
enable_annotations = os.environ.get("GATEWAY_ENABLE_ANNOTATIONS", "").lower() in ['true', '1']
enable_backed_mode = os.environ.get("GATEWAY_ENABLE_BACKED_MODE", "").lower() in ['true', '1']

env_vars = {
    "CELLXGENE_LOCATION": cellxgene_location,
    "CELLXGENE_DATA": cellxgene_data,
    "GATEWAY_IP": ip,
}

optional_env_vars = {
    "EXTERNAL_HOST": external_host,
    "EXTERNAL_PROTOCOL": external_protocol,
    "GATEWAY_PORT": gateway_port,
    "GATEWAY_EXTRA_SCRIPTS": extra_scripts,
    "GATEWAY_TTL": ttl,
    "GATEWAY_ENABLE_UPLOAD": enable_upload,
    "GATEWAY_ENABLE_ANNOTATIONS": enable_annotations,
    "GATEWAY_ENABLE_BACKED_MODE": enable_backed_mode,
}

def validate():
    if not all(env_vars.values()):
        raise ValueError(
            f"""
    Please ensure that environment variables are set correctly.
    The ones with None below are missing and need to be set.

    {env_vars}

    Set them at the terminal before running the gateway.
    An example is:

        export CELLXGENE_LOCATION=~/anaconda/envs/cellxgene-dev/bin/cellxgene
        export CELLXGENE_DATA=../cellxgene_data
        export GATEWAY_IP=127.0.0.1
    """
        )
    else:
        logging.getLogger("cellxgene_gateway").info(f"Got required env: {env_vars}", )
        logging.getLogger("cellxgene_gateway").info(f"Got optional env: {optional_env_vars}")



import os
import codecs
from setuptools import find_packages, setup
import sys

if sys.version_info < (3,6):
    sys.exit('Sorry, Python < 3.6 is not supported')

def read(rel_path):
    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, rel_path), 'r') as fp:
        return fp.read()

def get_version(rel_path):
    for line in read(rel_path).splitlines():
        if line.startswith('__version__'):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    else:
        raise RuntimeError("Unable to find version string.")

def parse_requirements():
    reqs = []
    with open("requirements.txt", "r") as f:
        for l in f.readlines():
            reqs.append(l.strip("\n"))
    return reqs

with open("README.md", "r") as fh:
    long_description = fh.read()

install_reqs = parse_requirements()

setup(
    # mandatory
    name="cellxgene-gateway",
    # mandatory
    version=get_version("cellxgene_gateway/__init__.py"),
    # mandatory
    author="Niket Patel, Yohann Potier, Alok Saldanha",
    author_email="alok.saldanha@novartis.com",
    description=("Cellxgene Gateway"),
    long_description=long_description,
    long_description_content_type="text/markdown",
    license="MIT",
    keywords="visualization, genomics",
    url="http://github.com/Novartis/cellxgene-gateway",
    packages=["cellxgene_gateway"],
    package_data={
        "cellxgene_gateway": [
            "static/css/homepagestyle.css",
            "static/nibr.ico",
            "templates/*.html"
    ]},
    data_files=[('', ['README.md', 'LICENSE'])],
    install_requires=install_reqs,
    entry_points={
        "console_scripts": ["cellxgene-gateway=cellxgene_gateway.gateway:main"]
    },
    classifiers=["Topic :: Scientific/Engineering :: Visualization"],
    python_requires='>=3.6',
)



# Copyright 2019 Novartis Institutes for BioMedical Research Inc. Licensed
# under the Apache License, Version 2.0 (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy
# of the License at http://www.apache.org/licenses/LICENSE-2.0. Unless
# required by applicable law or agreed to in writing, software distributed
# under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES
# OR CONDITIONS OF ANY KIND, either express or implied. See the License for
# the specific language governing permissions and limitations under the License.

__version__ = "0.2.0"



# Copyright 2019 Novartis Institutes for BioMedical Research Inc. Licensed
# under the Apache License, Version 2.0 (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy
# of the License at http://www.apache.org/licenses/LICENSE-2.0. Unless
# required by applicable law or agreed to in writing, software distributed
# under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES
# OR CONDITIONS OF ANY KIND, either express or implied. See the License for
# the specific language governing permissions and limitations under the License.

import time
import logging

from cellxgene_gateway.util import current_time_stamp
from cellxgene_gateway.env import ttl


class PruneProcessCache:
    def __init__(self, cache):
        self.cache = cache
        self.expire_seconds = (3600 if ttl is None else int(ttl))

    def __call__(self):
        while True:
            time.sleep(60)
            self.prune()

    def prune(self):
        timestamp = current_time_stamp()
        cutoff = timestamp - self.expire_seconds
        processes_to_delete = [p for p in self.cache.entry_list if  p.timestamp < cutoff]
        processes_to_keep = [p for p in self.cache.entry_list if not p.timestamp < cutoff]
        logger = logging.getLogger("cellxgene_gateway")
        logger.debug(f"Cutoff {cutoff} = timestamp {timestamp} - expire seconds {self.expire_seconds} , keeping {processes_to_keep}")
        
        for process in processes_to_delete:
            try:
                logger.info(f"pruning process {process.pid} ({process.key.dataset})")
                self.cache.prune(process)
            except Exception:
                logger.exception("failed to prune process {process.pid} ({process.dataset})")
